import io
import base64

import cv2
import numpy as np
from PIL import Image


class ImageTools:
    def __init__(self, src_or_bytes):
        # 转换为灰度图
        self.image = self.open_from_bytes(src_or_bytes) if isinstance(src_or_bytes, bytes) else src_or_bytes
        self.grayscale = cv2.cvtColor(self.image, cv2.COLOR_BGR2GRAY)
        self.h, self.w = self.grayscale.shape

    def area(self, src, thresh=None, gap_filling=(0, 0), erode_and_dilate=(0, 0)):
        """
        用于突出明暗变化密集的区域
        :param src: 灰度图
        :param thresh: 用于边缘检测的阈值
        :param gap_filling: 缝隙填充的像素
        :param erode_and_dilate: 腐蚀及膨胀的像素数
        :return: 二值化后的数组
        """
        src = cv2.Canny(src, thresh, 255)  # canny边缘检测
        if np.sum(gap_filling):
            src = self.gap_filling(src, *gap_filling)
        if np.sum(erode_and_dilate):
            src = self.erode_and_dilate(src, *erode_and_dilate)
        return src

    @staticmethod
    def binarization(src, thresh):
        """
        图片二值化为 0, 255 (黑, 白)并进行颜色反转
        :param src: 要处理的图像数组
        :param thresh: 阈值, 推荐 w // 15
        :return: 处理完的图像数组
        """
        _, src = cv2.threshold(src, thresh, 255, cv2.THRESH_BINARY_INV)
        return src

    @staticmethod
    def erode_and_dilate(src, erode, dilate):
        """
        腐蚀和膨胀
        :param src: 要处理的图像数组
        :param erode: 腐蚀的像素数
        :param dilate: 膨胀的像素数
        :return: 处理完的图像数组
        """
        src = cv2.erode(src, None, iterations=erode)
        src = cv2.dilate(src, None, iterations=dilate)
        return src

    @staticmethod
    def gap_filling(src, landscape, vertical):
        """
        消除图像缝隙
        :param src: 要处理的图像数组
        :param landscape: 横向填充的像素数
        :param vertical: 纵向填充的像素数
        :return: 处理完的图像数组
        """
        kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (landscape, vertical))
        src = cv2.morphologyEx(src, cv2.MORPH_CLOSE, kernel)
        return src

    @staticmethod
    def rotate_bound(src, angle, expand=True):
        """
        将图像逆时针旋转一定角度
        :param src: 要旋转的图片
        :param angle: 旋转角度
        :param expand: 是否保证原图的完整性, 否则保持像素不变, 旋转后角落内容丢失
        :return: 旋转后的图片
        """
        h, w = src.shape[:2]
        x_center, y_center = w / 2., h / 2.
        # 旋转中心, 旋转角度, 缩放比例
        m = cv2.getRotationMatrix2D((x_center, y_center), angle, 1)
        if not expand:
            src = cv2.warpAffine(src, m, (h, w))
            return src
        cos = np.abs(m[0, 0])
        sin = np.abs(m[0, 1])
        n_w = int((h * sin) + (w * cos))
        n_h = int((h * cos) + (w * sin))
        m[0, 2] += (n_w / 2) - x_center
        m[1, 2] += (n_h / 2) - y_center
        src = cv2.warpAffine(src, m, (n_w, n_h))
        return src

    def show_from_array(self, src):
        """
        图片显示
        :param src: 图片数组
        :return: None
        """
        self.array_to_image(src).show()

    @staticmethod
    def image_to_array(pil_img):
        """
        将PIL读取的图片转换为数组
        :param pil_img: PIL图片
        :return: ndarray
        """
        src = cv2.cvtColor(np.asarray(pil_img), cv2.COLOR_RGB2BGR)
        return src

    @staticmethod
    def array_to_image(src):
        """
        将数组转换为PIL格式的图片
        :param src: 图片数组, ndarray
        :return: PIL.Image
        """
        if src.ndim == 3:
            src = cv2.cvtColor(src, cv2.COLOR_BGR2RGB)
        return Image.fromarray(src)

    @staticmethod
    def open_from_bytes(byte, b64=True):
        """
        将字节码转换为图片数组
        :param byte: 图片字节码
        :param b64: 是否是base64编码的字节码
        :return: image
        """
        if b64:
            byte = base64.b64decode(byte)
        with io.BytesIO(byte) as f:
            image = Image.open(f)
            # 为了避免image离开with或函数将处于close状态
            src = __class__.image_to_array(image)
        return src

    @staticmethod
    def image_to_bytes(image, b64=True):
        """
        将图片转换为字节码
        :param image: image
        :param b64: 是否使用b64编码
        :return: 图片字节码
        """
        with io.BytesIO() as f:
            image.save(f, format="jpeg")
            byte = f.getvalue()
        if b64:
            byte = base64.b64encode(byte)
        return byte

    @staticmethod
    def point_rotate(point, center, angle):
        """
        图像(左上角为 0,0)上某个点point沿点center旋转angle角度后的坐标值
        :param point: 要旋转的点
        :param center: 旋转中心
        :param angle: 旋转角度
        :return: 旋转后的坐标
        """
        angle = -angle
        x = (point[0] - center[0]) * np.cos(np.pi / 180. * angle) \
            - (point[1] - center[1]) * np.sin(np.pi / 180. * angle) + center[0]
        y = (point[0] - center[0]) * np.sin(np.pi / 180. * angle) \
            + (point[1] - center[1]) * np.cos(np.pi / 180. * angle) + center[1]
        return x, y

    @staticmethod
    def rectangle(draw, left, top, width, height, angle=0, line_width=4, expand=2, fill="red"):
        """
        在图像上画一个矩形
        :param draw: 图像的可编辑对象 PIL.ImageDraw.Draw(img)
        :param left: 矩形左上角离图像左边的距离
        :param top: 矩形左上角离图像上边的距离
        :param width: 矩形宽度
        :param height: 矩形高度
        :param angle: 矩形旋转角度(沿矩形中心旋转)
        :param line_width: 线宽
        :param expand: 矩形向外扩大的像素数
        :param fill: 颜色
        :return: None
        """
        expand += line_width / 2.
        center = (left + width / 2., top + height / 2.)
        top_left = (left - expand, top - expand)
        bottom_left = (left - expand, top + height + expand)
        top_right = (left + width + expand, top - expand)
        bottom_right = (left + width + expand, top + height + expand)
        if angle != 0:
            top_left = __class__.point_rotate(top_left, center, angle)
            bottom_left = __class__.point_rotate(bottom_left, center, angle)
            top_right = __class__.point_rotate(top_right, center, angle)
            bottom_right = __class__.point_rotate(bottom_right, center, angle)
        draw.line([top_left, top_right],
                  fill=fill, width=line_width)  # 横线 1
        draw.line([bottom_left, bottom_right],
                  fill=fill, width=line_width)  # 横线 2
        draw.line([top_left, bottom_left],
                  fill=fill, width=line_width)  # 竖线 1
        draw.line([top_right, bottom_right],
                  fill=fill, width=line_width)  # 竖线 2



import os
import re

from .zxing import BarCodeReader


def qrdecode(qrcode_path, delete=True):
    result = None
    if os.name == 'nt':
        # 使用zbar识别二维码, 暂仅支持Windows
        zbarimg_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "ZBar_Win", "bin", "zbarimg.exe")
        barcode = os.popen(f"{zbarimg_path} -D {qrcode_path}").read()
        if barcode:
            result = barcode[8:]
    elif os.environ.get('JAVA_HOME'):
        # 使用zxing识别二维码, 基于Java环境
        try:
            zx = BarCodeReader()
            barcode = zx.decode(qrcode_path)
            if barcode:
                result = barcode.data
        except:
            pass
    else:
        # 使用php识别二维码, 基于php环境
        qrcode_decode_path = os.path.dirname(os.path.abspath(__file__)) + '/php_qrdecode'
        barcode = os.popen(f"php {qrcode_decode_path}/decode.php {qrcode_path}").read()
        if re.match(r"\d\d,\d\d", barcode):
            result = barcode
    if delete:
        os.remove(qrcode_path)
    return result



import abc
import json
import base64

import requests


class OCRBase:
    # 当请求失败时用于记录重复请求的次数
    _n = 0
    # 用于存放不同产品对应的url
    products = {}

    @abc.abstractmethod
    def set_header(self, **kwargs):
        """
        设置请求头
        :return: 请求头
        """
        pass

    @abc.abstractmethod
    def set_body(self, file_path, **kwargs):
        """
        设置body
        :param file_path: 文件路径
        :return: body
        """
        pass

    def get_access_token(self):
        """
        获取控制访问权限的加密串
        :return:
        """
        pass

    def get_url(self, product):
        """
        获取接口url
        :param product: 产品类型
        :return: url
        """
        return self.products.get(product, None)

    def ocr(self, file_path, product):
        """
        调用第三方接口
        :param file_path: 文件路径
        :param product: 请求的产品类型
        :return: 以字典形式返回第三方接口返回的内容
        """
        headers = self.set_header()
        body = self.set_body(file_path, product=product)
        url = self.get_url(product)
        if url is None:
            return {"error_code": "暂无此功能"}
        response = requests.post(url, data=body, headers=headers)
        content = json.loads(response.text)
        return content

    def get(self, file_path, product, code_key, right_value=None, abnormal_re_requests=3):
        """
        获取第三方返回值的入口
        :param file_path: 文件路径
        :param product: 产品类型
                        "handwriting": 手写体(暂时仅支持此值)
        :param code_key: 存放异常码的keys, 多层用 "." 隔开
        :param right_value: 请求正常的code值
        :param abnormal_re_requests: 请求失败后重新请求的次数(包括第一次)
        :return: 是否请求成功, 第三方返回的内容(dict)
        """
        content = self.ocr(file_path, product)
        if self.get_code(content, code_key) != right_value:
            self._n += 1
            if self._n >= abnormal_re_requests:
                return False, content
            else:
                self.get_access_token()
                return self.get(file_path, product, code_key, right_value, abnormal_re_requests)
        return True, content

    @staticmethod
    def get_code(content, code_key):
        """
        获取异常码, 无异常返回None, 或给定正常码
        :param content: 第三方接口的返回值(dict)
        :param code_key: 存放异常码的key
        :return: 返回异常码
        """
        keys = code_key.split('.')
        for i in keys:
            content = content.get(i, None)
            if content is None:
                return content
        return content

    @staticmethod
    def base64_img(img_path_or_bytes):
        """
        读取文件
        :param img_path_or_bytes: 文件路径
        :return: 返回经过base64编码的二进制字符串
        """
        if isinstance(img_path_or_bytes, bytes):
            return img_path_or_bytes
        with open(img_path_or_bytes, 'rb') as f:
            return base64.b64encode(f.read())



########################################################################
#
#  zxing.py -- a quick and dirty wrapper for zxing for python
#
#  this allows you to send images and get back data from the ZXing
#  library:  http://code.google.com/p/zxing/
#
#  by default, it will expect to be run from the zxing source code directory
#  otherwise you must specify the location as a parameter to the constructor
#
import os
import re
import subprocess

__version__ = '0.3'


class BarCodeReader:
    location = ""
    command = "java"
    libs = ["javase.jar", "core.jar"]
    args = ["-cp", "LIBS", "com.google.zxing.client.j2se.CommandLineRunner"]

    def __init__(self, loc=""):
        if not len(loc):
            if "ZXING_LIBRARY" in os.environ:
                loc = os.environ["ZXING_LIBRARY"]
            else:
                loc = os.path.dirname(os.path.abspath(__file__))

        self.location = loc

    def decode(self, files, try_harder=False, qr_only=False):
        cmd = [self.command]
        cmd += self.args[:]  # copy arg values
        if try_harder:
            cmd.append("--try_harder")
        if qr_only:
            cmd.append("--possibleFormats=QR_CODE")

        libraries = [self.location + "/" + l for l in self.libs]

        cmd = [c if c != "LIBS" else os.pathsep.join(libraries) for c in cmd]

        # send one file, or multiple files in a list
        single_file = False
        if not isinstance(files, list):
            cmd.append(files)
            single_file = True
        else:
            cmd += files

        (stdout, stderr) = subprocess.Popen(cmd, stdout=subprocess.PIPE, universal_newlines=True).communicate()
        codes = []
        file_results = stdout.split("\nfile:")
        for result in file_results:
            lines = stdout.split("\n")
            if re.search("No barcode found", lines[0]):
                codes.append(None)
                continue

            codes.append(BarCode(result))

        if single_file:
            return codes[0]
        else:
            return codes


# this is the barcode class which has
class BarCode:
    format = ""
    points = []
    data = ""
    raw = ""

    def __init__(self, zxing_output):
        lines = zxing_output.split("\n")
        raw_block = False
        parsed_block = False
        point_block = False

        self.points = []
        for ll in lines:
            m = re.search("format:\s([^,]+)", ll)
            if not raw_block and not parsed_block and not point_block and m:
                self.format = m.group(1)
                continue

            if not raw_block and not parsed_block and not point_block and ll == "Raw result:":
                raw_block = True
                continue

            if raw_block and ll != "Parsed result:":
                self.raw += ll + "\n"
                continue

            if raw_block and ll == "Parsed result:":
                raw_block = False
                parsed_block = True
                continue

            if parsed_block and not re.match("Found\s\d\sresult\spoints", ll):
                self.data += ll + "\n"
                continue

            if parsed_block and re.match("Found\s\d\sresult\spoints", ll):
                parsed_block = False
                point_block = True
                continue

            if point_block:
                m = re.search("Point\s(\d+):\s\(([\d.]+),([\d.]+)\)", ll)
                if m:
                    self.points.append((float(m.group(2)), float(m.group(3))))

        return


if __name__ == "__main__":
    print("ZXing module")



import cv2
import numpy as np
from PIL import ImageDraw

from .tool import ImageTools


class ImageProcess(ImageTools):
    @staticmethod
    def _near(data, nn=0, difference=5):
        """
        去除data中开头元素且相邻不超过difference的值
        比如[1, 2, 3, 4, 20, 21, 22]
        nn 跟 1 比较, 不超过 5 则 1 被剔除
        接下来依次, 2 跟 1 比较, 3 跟 2 比较
        直到 20 跟 4 比较大于 5 结束
        :param data: 一维有序数据结构
        :param nn: 首次判断时比较的值
        :param difference: 最大跨度
        :return: 过滤后的列表
        """
        data_copy = list(data)
        for ii in data:
            if -difference <= ii - nn <= difference:
                data_copy = data_copy[1:]
                nn = ii
            else:
                break
        return data_copy

    def correction(self, src, max_angle=5, threshold=100, angle=0.0):
        """
        对图片自动旋转矫正
        :param src: 要旋转的图片(二值, 背景黑色, 0, 255)
        :param max_angle: 允许的最大旋转角度
        :param threshold: 初始判断阈值, 当某一行的平均值小于此值将会继续进行
        :param angle: 初始旋转角度, 若指定则会先旋转此角度, 后进行继续判断
        :return: 旋转角度, 结束时阈值, 白色点最少的行
        """
        src_ = src
        if angle > max_angle or angle < -max_angle:
            return None, None, None
        if angle != 0:
            src = self.rotate_bound(src, angle)
        row_mean = src.mean(axis=1).round(2)
        row_min = row_mean.min()
        min_index = np.where(row_mean == row_min)[0]

        min_index = self._near(min_index, 0)
        min_index.reverse()
        min_index = self._near(min_index, src.shape[0])
        min_count = len(min_index)
        threshold_ = (row_min + 1) * 1. / (min_count + 1)
        if threshold_ <= threshold:
            result1 = result2 = 0.0, None, None
            if angle >= 0:
                result1 = self.correction(src_, threshold=threshold_, angle=angle + 0.5)
            if angle <= 0:
                result2 = self.correction(src_, threshold=threshold_, angle=angle - 0.5)
            if result1[1] is None:
                return (0.0, None, None) if result2[1] is None else result2
            else:
                if result2[1] is None:
                    return result1
                else:
                    return result1 if result1[1] < result2[1] else result2
        else:
            return angle, threshold_, min_index

    @staticmethod
    def _judging_the_edge(arr_1d, threshold=0, expansion=0):
        """
        返回去除两端小于阈值的索引范围
        :param arr_1d: 一维数组
        :param threshold: 阈值
        :param expansion: 扩展值(结果范围向两端扩展)
        :return: 索引开始值, 索引结束值
        """
        r = len(arr_1d)
        x1 = 0
        for x1 in range(r):
            if arr_1d[x1] > threshold:
                break
        x2 = r - 1
        for x2 in range(r - 1, 0, -1):
            if arr_1d[x2] > threshold:
                break
        x1 = 0 if x1 - expansion < 0 else x1 - expansion
        x2 = r if x2 + 1 + expansion > r else x2 + 1 + expansion
        return x1, x2

    def remove_the_edg(self, src, threshold=255 * 20, expansion=20):
        """
        去除二维矩阵边缘零值的索引范围
        :param src: 二维数组
        :param threshold: 阈值
        :param expansion: 扩展值(结果范围向两端扩展)
        :return: ((行开始, 行结束), (列开始, 列结束))
        """
        row = src.sum(axis=1)
        x1, x2 = self._judging_the_edge(row, threshold, expansion)
        column = src.sum(axis=0)
        y1, y2 = self._judging_the_edge(column, threshold, expansion)
        return (x1, x2), (y1, y2)


class SelectQrcode(ImageTools):
    @staticmethod
    def _select_area(array):
        """
        选择数组中有值的正方形区域
        :param array: 数组
        :return: 边长, 正方形位置(左上角)
        """
        index = list(zip(*np.where(array == 255)))
        if index:
            min_index = min(index, key=lambda x: x[0] + x[1])
            max_index = max(index, key=lambda x: x[0] + x[1])
            side_leng = ((max_index[0] - min_index[0]) + (max_index[1] - min_index[1])) // 2
            return side_leng, min_index[0], min_index[1]
        return 0, 0, 0

    def img_qrcode(self, src, show=False, binary_picture=False):
        """
        截取及标记二维码
        :param src: 灰度图
        :param show: 是否展示图片
        :param binary_picture: 是否展示处理过程图
        :return: 二维码图片数组, 及二维码位置
        """
        src_ = self.binarization(src, 40)

        ll = int(np.sqrt(self.h * self.w / 10000) + 5)
        src = self.area(src_, 30, gap_filling=(ll, ll))

        # 腐蚀和膨胀
        n = 0
        while True:
            src = cv2.erode(src, None, iterations=5)
            n += 1
            if n > 5 and src.sum() < 255 * 20 ** 2:
                break
        src = cv2.dilate(src, None, iterations=5 * n + 5)

        if binary_picture:
            self.show_from_array(src)

        sid_length, x, y = self._select_area(src)

        image_3 = self.array_to_image(self.image)

        position = [y, x, sid_length, sid_length]
        qrcode = False
        if sid_length:
            if src_[x - 10:x + sid_length + 10, y - 10:y + sid_length + 10].mean() > 255 * 0.1:
                qrcode = self.image[x - 10:x + sid_length + 10, y - 10:y + sid_length + 10]
        if show:
            draw = ImageDraw.Draw(image_3)
            self.rectangle(draw, *position, expand=7)
            image_3.show()
        return qrcode, position


class Segmentation(ImageTools):
    """
    未使用, 暂留
    """
    @staticmethod
    def is_blank(src, coordinate, is_img=False, threshold=0.00005, multiple=5):
        x1, x2, y1, y2 = coordinate
        if is_img:
            if ((x2 - x1) // (y2 - y1) >= multiple) \
                    or ((y2 - y1) // (x2 - x1) >= multiple):
                return True
        array = src[x1:x2, y1:y2]
        return array.mean() < 255 * threshold

    def segmentation(self, src, coordinate, dividing_line=50, *, h=False, w=False):
        if h ^ w:
            raise ValueError("需要h与w有且只有一个为True")
        x1, x2, y1, y2 = coordinate
        _h, _w = src[x1:x2, y1:y2].shape
        h, w = _h * h, _w * w
        h_center, w_center = h // 2, w // 2
        for i in range(dividing_line, h_center+w_center, dividing_line):
            h_effective = i if h else 0
            w_effective = i if h else 0
            coordinate_ = (
                x1 + h_center - h_effective, x1 + h_center - h_effective + dividing_line,
                y1 + w_center - w_effective, y1 + w_center - w_effective + dividing_line)
            if self.is_blank(src, coordinate_):
                out = h_center + w_center - i
                break
            coordinate_ = (
                x1 + h_center + h_effective - dividing_line, x1 + h_center + h_effective,
                y1 + w_center + w_effective - dividing_line, y1 + w_center + w_effective)
            if self.is_blank(src, coordinate_):
                out = h_center + w_center + i
                break
        else:
            return () if self.is_blank(src, coordinate, True) else (coordinate,)
        if w == 0:
            coordinate_1, coordinate_2 = (x1, x1 + out, y1, y2), (x1 + out, x2, y1, y2)
        else:
            coordinate_1, coordinate_2 = (x1, x2, y1, y1 + out), (x1, x2, y1 + out, y2)
        if self.is_blank(src, coordinate_1, True):
            return self.segmentation(src, coordinate_2, dividing_line, h=h, w=w)
        if self.is_blank(src, coordinate_2, True):
            return self.segmentation(src, coordinate_1, dividing_line, h=h, w=w)
        result = (*self.segmentation(src, coordinate_1, dividing_line, h=h, w=w),
                  *self.segmentation(src, coordinate_2, dividing_line, h=h, w=w))
        return result

    def run(self, src):
        h, w = src.shape[:2]
        s = self.segmentation(src, (0, h, 0, w), h=True)
        for i in s:
            for i2 in self.segmentation(src, i, w=True):
                for j in self.segmentation(src, i2, h=True):
                    image_array = self.image[j[0]:j[1], j[2]:j[3]]
                    self.show_from_array(image_array)



xf_keys = {
    "data_key": "data.block.0.line",
    "word_key": "word.0.content",
    "left_key": "location.top_left.x",
    "top_key": "location.top_left.y",
    "right_key": "location.right_bottom.x",
    "bottom_key": "location.right_bottom.y"
}



import json

from ._ocr_base import OCRBase


class Aliyun(OCRBase):
    products = {
        # 通用文字识别
        "general": "https://tysbgpu.market.alicloudapi.com/api/predict/ocr_general",
        # 通用高精度版
        "advanced":"https://ocrapi-advanced.taobao.com/ocrservice/advanced",
        # 发票识别
        "invoice": "https://ocrapi-invoice.taobao.com/ocrservice/invoice",
        "handwriting": None
    }

    def __init__(self, appcode):
        """
        :param appcode: 阿里云提供的appcode值
        """
        self.appcode = appcode

    def set_header(self, **kwargs):
        headers = {
            'Authorization': 'APPCODE ' + self.appcode,
            'Content-Type': 'application/json; charset=UTF-8'
        }
        return headers

    def set_body(self, file_path, **kwargs):
        product = kwargs['product']
        body = dict()
        image = "image" if product == "general" else "img"
        body[image] = self.base64_img(file_path).decode("utf-8")
        if product == "general":
            body["configure"] = {"main_size": 16, "output_prob": True}
        if product == "advanced":
            body["prob"] = "true"  # 显示每一行的置信度
            body["charInfo"] = "false"  # 单字识别
            body["rotate"] = "true"  # 自动旋转
            body["table"] = "false"  # 表格识别
        return json.dumps(body)



class Tool:
    __conf = {
        "壹": 1, "贰": 2, "叁": 3,
        "肆": 4, "伍": 5, "陆": 6,
        "柒": 7, "捌": 8, "玖": 9, "零": 0,
        "拾": 10, "佰": 100, "仟": 1000,
        "万": 10000, "亿": 100000000,
        "圆": 1, "角": 0.1, "分": 0.01
    }

    @staticmethod
    def get(dict_or_list, key):
        """
        获取对应键或索引的值
        :param dict_or_list: dict/list
        :param key: 索引/键
        :return: value
        """
        value = None
        if isinstance(dict_or_list, dict):
            if key[0] == "_" == key[-1] and key[1:-1].isnumeric():
                key = int(key[1:-1])
            value = dict_or_list.get(key, None)
        elif isinstance(dict_or_list, list):
            if key.isnumeric():
                index = int(key)
                if len(dict_or_list) > index:
                    value = dict_or_list[index]
        return value

    @staticmethod
    def get_value(dict_or_list, keys):
        """
        获取字典中对应key(多层嵌套)的值
        :param dict_or_list: 嵌套字典或列表
        :param keys: 多个key使用 "." 隔开
                     字典中想使用数字索引则需在两端加上 "_", 如 "_2_"
        :return: 按给定key取到的值
        """
        if keys is None:
            return None
        keys = keys.split(".")
        for key in keys:
            dict_or_list = __class__.get(dict_or_list, key)
            if dict_or_list is None:
                break
        return dict_or_list

    @staticmethod
    def __conversion(string):
        value = __class__.__conf[string[0]]
        for i in string[1:]:
            value *= __class__.__conf[i]
        return value

    @staticmethod
    def amount_conversion(string, level="圆"):
        """
        大写数字转换为数值
        :param string: 大写数字
        :param level: 起始单位, 只能是 "圆角分"
        :return: 转换结果
        """
        if string == "":
            return 0

        values = []
        for n, i in enumerate(string):
            if i in "圆角分":
                level = i
                string = string[n+1:]
                break
            v = __class__.__conf[i]
            if v < 10 or len(values) == 0 or len(values[-1]) == 2:
                values.append(i)
            else:
                values[-1] += i
        else:
            string = ""

        if not values:
            return 0

        value = __class__.__conversion(values[0])
        for i in values[1:]:
            v = __class__.__conversion(i)
            if v > value:
                value *= v
            else:
                value += v
        _level = "圆角分零零"["圆角分零".find(level)+1]
        return value * __class__.__conf[level] + __class__.amount_conversion(string, _level)



import json
import requests

from ._ocr_base import OCRBase


class Baidu(OCRBase):
    __url = "https://aip.baidubce.com/rest/2.0/ocr/v1"
    products = {
        # 二维码识别
        "qrcode": f"{__url}/qrcode",
        # 发票识别
        "vat_invoice": f"{__url}/vat_invoice",
        # 通用文字识别(高精度, 含位置信息)
        "accurate": f"{__url}/accurate",
        # 手写体识别
        "handwriting": f"{__url}/handwriting"
    }

    def __init__(self, client_id, client_secret):
        """
        值由百度提供
        :param client_id:
        :param client_secret:
        """
        self.client_id = client_id
        self.client_secret = client_secret
        self.access_token = self.get_access_token()

    def get_access_token(self):
        url = "https://aip.baidubce.com/oauth/2.0/token"
        grant_type = "client_credentials"
        url = f"{url}?grant_type={grant_type}&client_id={self.client_id}&client_secret={self.client_secret}"
        headers = {'Content-Type': 'application/json; charset=UTF-8'}
        response = requests.get(url, headers=headers)
        content = json.loads(response.text)
        return content["access_token"]

    def get_url(self, product):
        return f'{self.products[product]}?access_token={self.access_token}'

    def set_header(self, **kwargs):
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        return headers

    def set_body(self, file_path, **kwargs):
        img = self.base64_img(file_path)
        return {"image": img}



from PIL import Image

from .nlp import block
from .nlp.setting import xf_keys
from .image import ImageProcess
from .ocr_api import XunFei


def run(b64, adjust=True, **kwargs):
    process = ImageProcess(b64)
    img = process.array_to_image(process.image)
    if adjust:
        # 二值化
        w = process.w
        binarization = process.area(process.grayscale, w//15, gap_filling=(w//340, w//340), erode_and_dilate=(3, 3))
        # 去除边缘
        x, y = process.remove_the_edg(binarization, 0, 0)
        if binarization[x[0]:x[1], y[0]:y[1]].mean() > 11:
            # 获取应旋转的角度
            angle_ = process.correction(binarization)[0]
            # 旋转原始图片
            if angle_:
                color = process.image[:, 0].mean(axis=0)
                color = tuple(int(i) for i in color)
                img = img.rotate(angle_, Image.BICUBIC, expand=True, fillcolor=color)
                # img.show()
    # 将图片转换为b64编码的字节码
    b64 = process.image_to_bytes(img)

    # 调用讯飞的手写文字识别接口
    xf = XunFei(appid=kwargs["appid"], api_key=kwargs["api_key"])
    status, result = xf.get(b64, product="handwriting", code_key="code", right_value="0")

    if status:
        # 结果转换为DataFrame
        data_df = block.get_df(result, **xf_keys)
        # 返回分段结果
        string = block.block(data_df)
        if string:
            return string
        else:
            return result
    else:
        return result



import json
import time
import base64
import hashlib

from ._ocr_base import OCRBase


class XunFei(OCRBase):
    __url = "http://webapi.xfyun.cn/v1/service/v1/ocr"
    products = {
        "general": f"{__url}/general",
        "handwriting": f"{__url}/handwriting"
    }
    image = None

    def __init__(self, appid, api_key):
        """
        值由讯飞提供
        :param appid:
        :param api_key:
        """
        self.appid = appid
        self.api_key = api_key

    def set_header(self, **kwargs):
        headers = dict()
        headers["X-Appid"] = self.appid
        headers["X-CurTime"] = f"{int(time.time())}"
        param = json.dumps({"language": "cn|en", "location": "true"}).encode("utf8")
        headers["X-Param"] = base64.b64encode(param).decode("utf8")
        headers["X-CheckSum"] = f"{self.api_key}{headers['X-CurTime']}{headers['X-Param']}"
        h = hashlib.md5()
        h.update(headers["X-CheckSum"].encode("utf8"))
        headers["X-CheckSum"] = h.hexdigest()
        return headers

    def set_body(self, file_path, **kwargs):
        body = dict()
        self.image = self.base64_img(file_path).decode("utf8") if self.image is None else self.image
        body["image"] = self.image
        return body



import sys

from setuptools import setup, find_packages

import eoffcn_ai

if sys.version_info < (3, 5):
    sys.exit('Python 3.5 or greater is required.')

with open('README.rst', encoding="utf8") as fp:
    readme = fp.read()

setup(
    name='eoffcn_ai',
    version=eoffcn_ai.__version__,
    description='eoffcn ai api.',
    long_description=readme,
    author='PuKun',
    author_email='pukun@offcn.com',
    maintainer='PuKun',
    maintainer_email='pukun@offcn.com',
    packages=find_packages(),
    include_package_data=True,
    # package_data={'': ['*.jar', '*.php', '*.json', '*.bat', '*.dll', '*.exe']},
    install_requires=['numpy', 'opencv-python', 'pandas', 'pillow', 'requests'],
    platforms=['any'],
    classifiers=[]
    )



import pandas as pd

from ..tool import get_value


def filter_(df):
    return df


def get_df(dict_, data_key, word_key,
           left_key, top_key,
           right_key=None, bottom_key=None,
           height_key=None, width_key=None):
    """
    将字典转换为DataFrame
    :param dict_: 字典
    :param data_key: 获取行数据的key, "data.block.0.line"
    :param word_key: "word.0.content"
    :param left_key: "location.top_left.x"
    :param top_key: "location.top_left.y"
    :param right_key: "location.right_bottom.x"
    :param bottom_key: "location.right_bottom.y"
    :param height_key:
    :param width_key:
    :return: DataFrame
    """
    if right_key is None and width_key is None:
        raise ValueError("right_key与width_key至少有一个不为空")
    if bottom_key is None and height_key is None:
        raise ValueError("bottom_key与height_key至少有一个不为空")
    data_df = pd.DataFrame(columns=['word', 'left', 'top', 'right', 'bottom', 'height', 'width'])
    data = get_value(dict_, data_key)
    for n, row in enumerate(data):
        data_df.at[n, 'word'] = get_value(row, word_key)
        data_df.at[n, 'left'] = get_value(row, left_key)
        data_df.at[n, 'top'] = get_value(row, top_key)
        data_df.at[n, 'right'] = get_value(row, right_key)
        data_df.at[n, 'bottom'] = get_value(row, bottom_key)
        data_df.at[n, 'height'] = get_value(row, height_key)
        data_df.at[n, 'width'] = get_value(row, width_key)
        if data_df.at[n, 'height'] is None:
            data_df.at[n, 'height'] = data_df.at[n, 'bottom'] - data_df.at[n, 'top']
        else:
            data_df.at[n, 'bottom'] = data_df.at[n, 'top'] + data_df.at[n, 'height']
        if data_df.at[n, 'width'] is None:
            data_df.at[n, 'width'] = data_df.at[n, 'right'] - data_df.at[n, 'left']
        else:
            data_df.at[n, 'right'] = data_df.at[n, 'left'] + data_df.at[n, 'width']
    return data_df


def block(df, left=80, scope=(80, 200), filter_height=0.7):
    """
    分段
    :param df: get_df的返回值
    :param left: 求所有左侧坐标小于此值的平均值
    :param scope: 用于判断段首的位置, 在此范围内将被判断为段首
    :param filter_height: 行高小于 平均行高*filter_height 的行将被过滤
    :return: 含换行符的string
    """
    df = filter_(df)
    height_mean = df['height'].mean()
    df = df[df['height'] > height_mean * filter_height]
    left_mean = df['left'][df['left'] < left].mean()
    paragraph = (df['left'] > left_mean + scope[0]) & (df['left'] < left_mean + scope[1])
    paragraph_index = paragraph[paragraph].index
    for i in paragraph_index:
        df.at[i, 'word'] = '\n  ' + df.at[i, 'word']
    return df['word'].str.cat(sep='')


if __name__ == "__main__":
    print(get_value({"a": [1, {"b": 2}]}, "a.1.b"))
